/**
 * Quad Sim v0.21
 * Copyright (C) 2005  Giliam de Carpentier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * You can contact me at giliam@decarpentier.nl for questions concerning this license.
 */
#include "common.h"
// ----------------------------------------------------------------------------
// CLASS: MB (Mass Body)
// DESCR: simulates a spring between two mass points
// ----------------------------------------------------------------------------
MB::MB(float m, const V &vTheta) : MP(m) {
	this->vTheta = vTheta; vM = VZERO;
}

MB::MB(const PVV &pvv, const PVV &pvvRot, float m, const V &vTheta) : MP(&pvv, m) {
	rmPhiT = !(rmPhi = RM(pvvRot.vP));
	rmDelta = RM(vOmega = pvvRot.vV * TAU);
	vM = VZERO;	this->vTheta = vTheta;
}

V MB::w2m(const V &v)		{return rmPhiT * (v - vP);}

V MB::m2w(const V &v)		{return (rmPhi * v) + vP;}

PVV MB::m2pvv(const V &v)	{return PVV(m2w(v), rmPhi * (vOmega ^ v) + vV);}

void MB::addMoment(const V &vM) {this->vM += vM;}

void MB::addForce(const V &vF) {this->vF += vF;}

void MB::addForce(const V &vP, const V &vF) {
	this->vF += vF;	vM += (vP - this->vP) ^ vF;
}

void MB::advance() {
	MP::advance();
	rmDelta *= RM(rmPhiT * vM / vTheta * (TAU * TAU));
	vOmega = rmDelta.vector() / TAU; 
	rmPhiT = !(rmPhi *= rmDelta); vM = VZERO;
}

// ----------------------------------------------------------------------------
// CLASS: RMB (Relative Mass Body) implementation
// DESCR: defines a child MB relative to a parent MB
// ----------------------------------------------------------------------------
RMB::RMB(MB *pmb, const V &vRelP) : MB(pmb->m, pmb->vTheta) {
	this->pmb = pmb; this->vRelP = vRelP; V vPP = vRelP * vRelP;
	vTheta += m * V(vPP.y + vPP.z, vPP.z + vPP.x, vPP.x + vPP.y);
	post();	
}

PVV RMB::m2pvv(const V &v)	{
	return pmb->m2pvv(vRelP + v);
}

void RMB::pre() {
	pmb->addForce(vP, vF);
	pmb->addMoment(vM);
}

void RMB::advance() {
	vM = vF = VZERO;
}

void RMB::post() {
	rmPhi = pmb->rmPhi; rmPhiT = pmb->rmPhiT;
	rmDelta = pmb->rmDelta; vOmega = pmb->vOmega;
	PVV pvv = pmb->m2pvv(vRelP); vV = pvv.vV; vP = pvv.vP;
}

// ----------------------------------------------------------------------------
// CLASS: SPRING implementation
// DESCR: simulates a spring between two mass points
// ----------------------------------------------------------------------------
SPRINGPP::SPRINGPP(MP *pmpA, MP *pmpB, float fS, float fD) {
	ppvvA = pmpA; ppvvB = pmpB;
	init(1.0f / (1.0f / pmpA->m + 1.0f / pmpB->m), fS, fD);
}

void SPRINGPP::init(float max, float fS, float fD) {
	length = (ppvvB->vP - ppvvA->vP).len(); 
	cS = fS * max / TAU / TAU; cD = fD * max / TAU;
}


V SPRINGPP::force() {
	V vDelta = ppvvB->vP - ppvvA->vP; float delta = vDelta.len();
	if (delta == 0.0f) return VZERO; 
	float force = cS * (length - delta) -
				  cD / delta * ((ppvvB->vV - ppvvA->vV) & vDelta);
	return vDelta * (force / delta);
}

void SPRINGPP::addForce(const V &vF) {
	((MP *)ppvvA)->addForce(-vF);
	((MP *)ppvvB)->addForce(vF);
}


// ----------------------------------------------------------------------------
// CLASS: SPRINGBP
// DESCR: simulates a spring between a mass body and a mass point
// ----------------------------------------------------------------------------
SPRINGBP::SPRINGBP(MB *pmbA, const V &vA, MP *pmpB, float fS, float fD) {
	this->pmbA = pmbA; this->vA = vA; ppvvA = new PVV();
	*ppvvA = pmbA->m2pvv(vA); ppvvB = pmpB;	V vAA = vA * vA;
	init(1.0f / (1.0f / pmbA->m + 1.0f / pmpB->m + (V(vAA.y + vAA.z,
		vAA.z + vAA.x, vAA.x + vAA.y) / pmbA->vTheta).largest()), fS, fD);
}

SPRINGBP::~SPRINGBP() {
	delete ppvvA;
}

V SPRINGBP::force() {
	*ppvvA = pmbA->m2pvv(vA);
	return SPRINGPP::force();
}

void SPRINGBP::addForce(const V &vF) {
	pmbA->addForce(ppvvA->vP, -vF);
	((MP *)ppvvB)->addForce(vF);
}

// ----------------------------------------------------------------------------
// CLASS: SPRINGBB
// DESCR: simulates a spring between a mass body and a mass point
// ----------------------------------------------------------------------------
SPRINGBB::SPRINGBB(MB *pmbA, const V &vA, MB *pmbB, const V &vB, float fS, float fD) {
	this->pmbA = pmbA; this->vA = vA; ppvvA = new PVV(); *ppvvA = pmbA->m2pvv(vA); 
	this->pmbB = pmbB; this->vB = vB; ppvvB = new PVV(); *ppvvB = pmbB->m2pvv(vB);
	V vAA = vA * vA, vBB = vB * vB;
	init(1.0f / (1.0f / pmbA->m + 1.0f / pmbB->m + (V(vAA.y + vAA.z, vAA.z + vAA.x,
		vAA.x + vAA.y) / pmbA->vTheta).largest() + (V(vBB.y + vBB.z, vBB.z + vBB.x,
		vBB.x + vBB.y) / pmbB->vTheta).largest()), fS, fD);
}

SPRINGBB::~SPRINGBB() {
	delete ppvvA, ppvvB;
}

V SPRINGBB::force() {
	*ppvvA = pmbA->m2pvv(vA); *ppvvB = pmbB->m2pvv(vB);
	return SPRINGPP::force();
}

void SPRINGBB::addForce(const V &vF) {
	pmbA->addForce(ppvvA->vP, -vF);
	pmbB->addForce(ppvvB->vP, vF);
}

