/**
 * Quad Sim v0.21
 * Copyright (C) 2005  Giliam de Carpentier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * You can contact me at giliam@decarpentier.nl for questions concerning this license.
 */
#include "common.h"

V vTest;//////////

// ----------------------------------------------------------------------------
// CLASS: AXLE implementation
// DESCR: superclass that controls an axle coupled to a parent MB
// ----------------------------------------------------------------------------
AXLE::AXLE(MB *pmb, float theta, float relPhi, float relOmega) :
	MB(pmb->m, pmb->vTheta) {
	this->relPhi = relPhi; this->relOmega = relOmega; this->pmb = pmb;
	vTheta.x = theta; relAlpha = 0.f; oldParentOmega = pmb->vOmega.x; post(); 
}

void AXLE::addFriction(float max) {
	float ta = vTheta.x, tp = pmb->vTheta.x;
	float moment = ((rmPhiT*(tp*vM-ta*pmb->vM)).x+relOmega*ta*tp/TAU)/(ta+tp);
	moment = (moment > 0 ? MIN(moment, max) : MAX(moment, -max));
	V v = rmPhi * V(moment, 0, 0); addMoment(-v); pmb->addMoment(v);
}

void AXLE::pre() {
	V intM = rmPhiT * vM;
	relAlpha = intM.x / vTheta.x + (oldParentOmega - pmb->vOmega.x) / TAU;
	pmb->addForce(vF);
	pmb->addMoment(rmPhi * (V(0, 1, 1) * intM));
	oldParentOmega = pmb->vOmega.x;
}

void AXLE::advance() {
	relPhi += (relOmega += relAlpha * TAU) * TAU;
	if (relPhi > PI) do relPhi -= TWOPI; while (relPhi > PI);
	else while (relPhi < -PI) relPhi += TWOPI;
	vF = vM = VZERO;
}

void AXLE::post() {
	rmPhiT = !(rmPhi = pmb->rmPhi * RM(V(relPhi, 0, 0)));
	rmDelta = RM(rmPhiT * pmb->rmPhi * pmb->vOmega * TAU) * 
		      RM(V(relOmega * TAU, 0, 0));
	vOmega = rmDelta.vector() / TAU; vV = pmb->vV; vP = pmb->vP;
}

// ----------------------------------------------------------------------------
// CLASS: WHEEL declaration
// DESCR: defines a WHEEL based on MPs coupled to an AXLE
//        can add MPs forces to AXLE and render the WHEEL
// ----------------------------------------------------------------------------
WHEEL::WHEEL(AXLE *paxle, int n, const V &vC, float r, float w, float mpm,
			 float cS, float cD, float cFric, float cFric0, const V &vProf) {
	this->paxle = paxle; this->n = n; this->vC = vC; this->r = r; this->w = w;
	this->cFric = cFric; this->cFric0 = cFric0;	cFricProf = ((V)vProf).len(); 
	this->vProf = vProf / cFricProf; float ang = TWOPI / n; gripFact = 0.f;
	ppmp = new MP*[n]; ppspring = new SPRINGBP*[n];
	for (int i = 0; i < n; i++) {
		V v = vC + V(w * ((i & 1) - 0.5f), r * sinf(i * ang), r * cosf(i * ang));
		ppmp[i] = new MP(&paxle->m2pvv(v), mpm);
		ppspring[i] = new SPRINGBP(paxle, v, ppmp[i], cS, cD);
	}
}

WHEEL::~WHEEL() {
	for (int i = 0; i < n; i++) {
		delete ppspring[i], ppmp[i];
	}
	delete ppspring, ppmp;
}

void WHEEL::pre() {
	float angStep = TWOPI / n;
	float ang = -paxle->relPhi;

	V vCenter = paxle->m2w(vC);
	V vProfW = paxle->rmPhi * vProf;

	vGripTot = VZERO;
	for (int i = 0; i < n; i++) {
		ppspring[i]->addForce(ppspring[i]->force());
		ppmp[i]->addForce(gravity.force(ppmp[i]));
		ppmp[i]->addForce((vCenter - ppmp[i]->vP) * ppmp[i]->m * paxle->relOmega * paxle->relOmega);
		//if (ang < 0) 
			ppmp[i]->addForce(map.force(ppmp[i], cFric, cFric0, vProfW, cFricProf));
		ang += angStep; if (ang > PI) ang -= TWOPI;
	}
	FILTER(gripFact, MAX(0.f, vGripTot.len() * .0002f - .5f), .99f);
}

void WHEEL::advance() {
	for (int i = 0; i < n; i++)
		ppmp[i]->advance();
}

void WHEEL::post() {
}

void WHEEL::render() {
	int i;
	
	glBegin(GL_TRIANGLE_FAN);
		glColor3f(0.2f+gripFact,0.18f,0.15f);
		glVertex3fv(paxle->m2w(vC + V(w * -.5f, 0.f, 0.f)).p);
		for (i = 0; i < n; i += 2)
			glVertex3fv(paxle->m2w(ppspring[i]->vA * V(1, 0.92f, 0.92f)).p);
		glColor3f(0.6f,0.6f,0.55f);
		glVertex3fv(paxle->m2w(ppspring[0]->vA * V(1, 0.92f, 0.92f)).p);
	glEnd();

	glBegin(GL_TRIANGLE_FAN);
		glColor3f(0.2f+gripFact,0.18f,0.15f);
		glVertex3fv(paxle->m2w(vC + V(w * .5f, 0.f, 0.f)).p);
		for (i = n - 2; i >= 0; i -= 2)
			glVertex3fv(paxle->m2w(ppspring[i]->vA * V(1, 0.92f, 0.92f) + V(w,0,0)).p);
		glColor3f(0.6f,0.6f,0.55f);
		glVertex3fv(paxle->m2w(ppspring[n-2]->vA * V(1, 0.92f, 0.92f) + V(w,0,0)).p);
	glEnd();

	
	V v;
	glBegin(GL_TRIANGLE_STRIP);
		glColor3f(0.1f,0.08f,0.04f);
		for (i = 0; i <= n; i += 2) {
			glVertex3fv(paxle->m2w(ppspring[i % n]->vA * V(1, 0.92f, 0.92f)).p);
			glVertex3fv(paxle->m2w(ppspring[(i + 1) % n]->vA * V(0,1,1) + V(vC.x,0,0)).p);
		}
	glEnd();

	glBegin(GL_TRIANGLE_STRIP);
		glColor3f(0.1f,0.08f,0.04f);
		for (i = n; i >= 0; i -= 2) {
			glVertex3fv(paxle->m2w(ppspring[i % n]->vA * V(1, 0.92f, 0.92f) + V(w,0,0)).p);
			glVertex3fv(paxle->m2w(ppspring[(i + n - 1) % n]->vA * V(0,1,1) + V(vC.x,0,0)).p);
		}
	glEnd();

}

// ----------------------------------------------------------------------------
// CLASS: BSS (Back Sub-System) implementation
// DESCR: defines the back swing arm, axle and wheels
// ----------------------------------------------------------------------------
extern float handbrake, leftbrake;


BSS::BSS(QUAD *pquad) {
	this->pquad = pquad;
	V vArm = V(0, .05f, .26f), vBearing = V(0, -.22f, .68f);
	pbearing = new MB(pquad->pmb->m2pvv(vBearing),
					  PVV(pquad->pmb->rmPhi.vector(), pquad->pmb->vOmega),
					  20.f, pquad->pmb->vTheta);
	paxle = new AXLE(pbearing, 1.2f, 0, 0);
//	pwheelL = new WHEEL(paxle, 24, V(-.45f,0.f,0.f), .22f, .24f, .22f, .14f, 0.99f, .5f, .4f, V(0,0,.08f));
//	pwheelR = new WHEEL(paxle, 24, V(.45f,0.f,0.f), .22f, .24f, .22f, .14f, 0.99f, .5f, .4f, V(0,0,.08f));
	pwheelL = new WHEEL(paxle, 24, V(-.45f,0.f,0.f), .22f, .24f, .22f, .04f, 0.99f, .4f, .5f, V(0,0,.05f));
	pwheelR = new WHEEL(paxle, 24, V(.45f,0.f,0.f), .22f, .24f, .22f, .04f, 0.99f, .4f, .5f, V(.15f,0,.05f));
	parm = new SPRINGBB(pquad->pmb, vArm, pbearing, VZERO, .8f, .75f);
//	pshock = new SPRINGBB(pquad->pmb, vBearing, pbearing, VZERO, .0022f, .3f);
	pshock = new SPRINGBB(pquad->pmb, vBearing, pbearing, VZERO, .0012f, .04f);
}

BSS::~BSS() {
	delete parm, pshock, pwheelL, pwheelR, paxle, pbearing;
}

void BSS::pre() {
	pwheelL->pre(); pwheelR->pre(); 
	paxle->addFriction(50 + handbrake * 600); paxle->pre();
	pbearing->addForce(gravity.force(pbearing));
	parm->addForce(parm->force());
	pshock->addForce(pshock->force());
	float moment = pquad->pmb->rmPhi.vA & pbearing->vM;
	V vD = parm->ppvvB->vP - parm->ppvvA->vP;
	V vF = moment * (vD ^ pquad->pmb->rmPhi.vA) / (vD & vD);
	pbearing->addForce(-vF); pquad->pmb->addForce(parm->ppvvA->vP, vF);
	pquad->pmb->addForce(pbearing->vP, 
		(pquad->pmb->rmPhi.vA & pbearing->vF) * pquad->pmb->rmPhi.vA);
	pquad->pmb->addMoment(pbearing->vM - moment * pquad->pmb->rmPhi.vA);	
}

void BSS::advance() {
	pwheelL->advance(); pwheelR->advance(); paxle->advance(); pbearing->advance();
}

void BSS::post() {
	PVV pvv = pquad->pmb->m2pvv(pquad->pmb->w2m(pbearing->vP) * V(0, 1, 1));
	pbearing->vP = pvv.vP;
	pbearing->vV += (pquad->pmb->rmPhi.vA & (pvv.vV - pbearing->vV)) * pquad->pmb->rmPhi.vA;
	pbearing->vOmega = pquad->pmb->vOmega; pbearing->rmDelta = pquad->pmb->rmDelta;
	pbearing->rmPhi = pquad->pmb->rmPhi; pbearing->rmPhiT = pquad->pmb->rmPhiT;
	paxle->post(); pwheelL->post(); pwheelR->post();
}

void BSS::render() {
	glBegin(GL_LINES);
		glColor3f(0.f, 0.f, 0.f);
		glVertex3fv(paxle->m2w(pwheelL->vC).p);
		glVertex3fv(paxle->m2w(pwheelR->vC).p);
	glEnd();

	V vP[8];
	V vB = pquad->pmb->w2m(pbearing->vP);
	V vN = ((vB - parm->vA) ^ V(1,0,0)).normal() * .075f;
	for (int i = 0; i < 8; i++)
		vP[i] = pquad->pmb->m2w(V((i & 1 ? .12f : -.12f), 0, 0) +
						   (i & 2 ? vN : -vN) +
						   (i & 4 ? vB : parm->vA));

	glBegin(GL_QUADS);
		glColor3f(0.6f,0.6f,0.0f);
		glVertex3fv(vP[1].p); glVertex3fv(vP[0].p); glVertex3fv(vP[2].p); glVertex3fv(vP[3].p);
		glVertex3fv(vP[4].p); glVertex3fv(vP[5].p); glVertex3fv(vP[7].p); glVertex3fv(vP[6].p);

		glColor3f(0.0f,0.6f,0.6f);
		glVertex3fv(vP[2].p); glVertex3fv(vP[0].p); glVertex3fv(vP[4].p); glVertex3fv(vP[6].p);
		glVertex3fv(vP[5].p); glVertex3fv(vP[1].p); glVertex3fv(vP[3].p); glVertex3fv(vP[7].p);

		glColor3f(0.6f,0.0f,0.6f);
		glVertex3fv(vP[0].p); glVertex3fv(vP[1].p); glVertex3fv(vP[5].p); glVertex3fv(vP[4].p);
		glVertex3fv(vP[3].p); glVertex3fv(vP[2].p); glVertex3fv(vP[6].p); glVertex3fv(vP[7].p);
	glEnd();


	pwheelL->render(); pwheelR->render();

}

// ----------------------------------------------------------------------------
// CLASS: FSS (Front Sub-System) implementation
// DESCR: defines the back swing arm, axle and wheels
// ----------------------------------------------------------------------------
FSS::FSS(QUAD *pquad, bool left) {
	this->pquad = pquad; side = (left ? 1.f : -1.f);
	V vArm = V(side * .08f, -.03f, -.46f), vBearing = V(side * .32f, -.22f, -.50f);
	rmCasterDelta = RM(V(atanf((vBearing.z-vArm.z)/(vBearing.y-vArm.y)),0,0)); 
	casterDist = rmCasterDelta.vC & vBearing; vSteerN = rmCasterDelta * V(0,1,0);
	rmCaster = pquad->pmb->rmPhi * rmCasterDelta;
	pbearing = new MB(pquad->pmb->m2pvv(vBearing),
					  PVV(pquad->pmb->rmPhi.vector(), pquad->pmb->vOmega),
					  8.f, pquad->pmb->vTheta);
	paxle = new AXLE(pbearing, .45f, 0, 0);
//	pwheel = new WHEEL(paxle, 24, V(side * .09f,0,0), .23f, .17f, .21f, .14f, 0.99f, .45f, .7f, V(.1f,0,0));
	pwheel = new WHEEL(paxle, 24, V(side * .09f,0,0), .23f, .17f, .21f, .04f, 0.99f, .4f, .5f, V(.2f,0,0));
	parm = new SPRINGBB(pquad->pmb, vArm, pbearing, VZERO, .8f, .75f);
//	pshock = new SPRINGBB(pquad->pmb, vBearing, pbearing, VZERO, .0032f, .3f);
	pshock = new SPRINGBB(pquad->pmb, vBearing, pbearing, VZERO, .0019f, .25f);
	post();
}

FSS::~FSS() {
	delete parm, pshock, pwheel, paxle, pbearing;
}

void FSS::pre() {
	pwheel->pre();
	paxle->addFriction(3);// + (side > .0f ? leftbrake * 300 : 0));
	paxle->pre();
	pbearing->addForce(gravity.force(pbearing));

	parm->addForce(parm->force());
	pshock->addForce(pshock->force());
	pquad->pmb->addForce(pbearing->vP, (rmCaster.vC & pbearing->vF) * rmCaster.vC);
	pquad->pmb->addMoment(pbearing->vM);
}

void FSS::advance() {
	pwheel->advance(); paxle->advance(); 
	pbearing->advance();
}

extern float steerAng;
void FSS::post() {
	V v = pquad->pmb->w2m(pbearing->vP);
	PVV pvv = pquad->pmb->m2pvv(v + (casterDist - (rmCasterDelta.vC & v)) * rmCasterDelta.vC);
	pbearing->vP = pvv.vP; rmCaster = pquad->pmb->rmPhi * rmCasterDelta;
	pbearing->vV += (rmCaster.vC & (pvv.vV - pbearing->vV)) * rmCaster.vC;

	RM rm = RM(vSteerN * atanf(1.0f / (1.0f / tanf(steerAng) - .33f * side)));
	pbearing->rmPhiT = !(pbearing->rmPhi = pquad->pmb->rmPhi * rm);
	pbearing->rmDelta = RM((pbearing->vOmega = !rm * pquad->pmb->vOmega) * TAU);
	paxle->post(); pwheel->post();
}

void FSS::render() {
	pwheel->render();
	glBegin(GL_LINES);
		glColor3f(.6f, .1f, .1f);
		glVertex3fv((parm->ppvvA->vP+rmCaster*V(0,.07f,.09f)).p);
		glVertex3fv((parm->ppvvB->vP+rmCaster*V(0,.07f,0)).p);
		glVertex3fv((parm->ppvvA->vP+rmCaster*V(0,.07f,-.09f)).p);
		glVertex3fv((parm->ppvvB->vP+rmCaster*V(0,.07f,0)).p);

		glVertex3fv((parm->ppvvA->vP+rmCaster*V(0,-.07f,.09f)).p);
		glVertex3fv((parm->ppvvB->vP+rmCaster*V(0,-.07f,0)).p);
		glVertex3fv((parm->ppvvA->vP+rmCaster*V(0,-.07f,-.09f)).p);
		glVertex3fv((parm->ppvvB->vP+rmCaster*V(0,-.07f,0)).p);

		glColor3f(0.f, 0.f, 0.f);
		glVertex3fv((parm->ppvvB->vP+rmCaster*V(0,.07f,0)).p);
		glVertex3fv((parm->ppvvB->vP+rmCaster*V(0,-.07f,0)).p);

	glEnd();

//	glBegin(GL_LINES);
//		glColor3f(0.f, 0.f, 1.f);
//		glVertex3fv(pbearing->vP.p);
//		glVertex3fv((pbearing->vP+rmCaster.vC).p);
//	glEnd();

}


// ----------------------------------------------------------------------------
// CLASS: QUAD implementation
// DESCR: defines a quad
// ----------------------------------------------------------------------------
QUAD::QUAD(MB *pmb) {
	this->pmb = pmb;
	pfssL = new FSS(this, true);
	pfssR = new FSS(this, false);
	pbss = new BSS(this);
}

void QUAD::pre() {
	pfssL->pre(); pfssR->pre();
	pbss->pre();
}

void QUAD::advance() {
	pmb->advance();
	pbss->advance();
	pfssL->advance(); pfssR->advance();
}

void QUAD::post() {
	pfssL->post(); pfssR->post();
	pbss->post();
}

void QUAD::render() {
	pfssL->render(); pfssR->render();
	pbss->render();
}










